/* $Id: client-gdbm.c,v 1.6 2001/06/23 00:50:29 robey Exp $
 *
 * Do gdbm queries to fetch/create users/channels, filling in Client entries.
 *
 * gdbm is quite a bit more primitive than SQL (which isn't exactly a pillar
 * of modernity itself) so we set our sights really low here.  Basically we
 * convert the client data to a string when writing it to the database, and
 * convert it back when reading.  We use two database files: one for users,
 * and one for channels.
 *
 * For stored messages, we use the destination id as the key, and the messages
 * (since we know they're enaml blocks) are clumped into one data block,
 * separated by '\n'.
 *
 * Copyright (C) 2000, 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <say2/say2.h>
#ifdef HAVE_GDBM
#include <gdbm.h>
#else
#include <ndbm.h>
#endif
#ifndef GDBM_SYNC
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include "client.h"
#include "client-db.h"
#include "client-gdbm.h"

#define BLOCK_SIZE	4096

#ifndef HAVE_GDBM
/* alias standard gdbm stuff into ndbm equivalents */
#define GDBM_FILE DBM *
#define gdbm_open(filename, blocksize, flags, mode, fatalfunc) \
	dbm_open(filename, flags, mode)
#define gdbm_fetch dbm_fetch
#define gdbm_store dbm_store
#define gdbm_close dbm_close
#define GDBM_INSERT DBM_INSERT
#define GDBM_REPLACE DBM_REPLACE
#define GDBM_WRITER O_RDWR|O_CREAT
#endif	/* HAVE_GDBM */

#ifndef GDBM_SYNC
/* GDBM 1.7 (not everyone's got 1.8, robey) doesn't have this */
#define GDBM_SYNC O_SYNC
#endif /* GDBM_SYNC */

typedef struct {
        GDBM_FILE userdb;
        GDBM_FILE chandb;
	GDBM_FILE msgdb;
} GDBMInfo;

/* convert client to an enaml string -- only done when creating a user */
static void
client_update_block (const Client *client, EnamlBlock *block)
{
        char *prefs, *privs;

        enaml_add_string (block, "id", client->id, 0);
        enaml_add_string (block, "email", client->email, 0);
        if (client->tagline != NULL) {
                enaml_add_string (block, "tagline", client->tagline, 0);
        } else {
                enaml_drop_key (block, "tagline");
        }
        prefs = client_prefs_to_string (client);
        enaml_add_string (block, "prefs", prefs, ENAML_COPY);
        g_free (prefs);
        if (client->icon != NULL) {
                enaml_add_blob (block, "icon", client->icon, client->iconsize, 0);
        } else {
                enaml_drop_key (block, "icon");
        }
        if (client->attributes != NULL) {
                enaml_add_block (block, "attributes", client->attributes, 0);
        } else {
                enaml_drop_key (block, "attributes");
        }

        if (client->flags & CLIENT_USER) {
                privs = client_privs_to_string (client);
                enaml_add_string (block, "privs", privs, ENAML_COPY);
                g_free (privs);
        }
}

static EnamlBlock *
read_id_block (GDBMInfo *info, const char *id, int flags)
{
        EnamlBlock *block;
        datum key, value;
        char *id_lower;

        id_lower = g_strdup (id);
        g_strdown (id_lower);
        key.dptr = id_lower;
        key.dsize = strlen (id_lower)+1;
        if (flags & CLIENT_USER) {
                value = gdbm_fetch (info->userdb, key);
        } else {
                value = gdbm_fetch (info->chandb, key);
        }
        g_free (id_lower);

        if (value.dptr == NULL) {
                return NULL;
        }
        block = enaml_parse (value.dptr);
        free (value.dptr);
        return block;
}

static EnamlBlock *
read_client_block (GDBMInfo *info, const Client *client)
{
        return read_id_block (info, client->id, client->flags);
}

static gboolean
write_client_block (GDBMInfo *info, const Client *client, EnamlBlock *block, gboolean exclusive)
{
        char *str;
        char *id_lower;
        datum key, value;
        int ret;

        str = enaml_compose (block, FALSE);
        id_lower = g_strdup (client->id);
        g_strdown (id_lower);
        key.dptr = id_lower;
        key.dsize = strlen (id_lower)+1;
        value.dptr = str;
        value.dsize = strlen (str)+1;

        if (client->flags & CLIENT_CHANNEL) {
                ret = gdbm_store (info->chandb, key, value, exclusive ? GDBM_INSERT : GDBM_REPLACE);
        } else {
                ret = gdbm_store (info->userdb, key, value, exclusive ? GDBM_INSERT : GDBM_REPLACE);
        }

        g_free (str);
        g_free (id_lower);
        return (ret == 0);
}

#ifdef HAVE_GDBM
static void
client_gdbm_fatal (const char *errmsg)
{
        log ("ERROR: gdbm: %s", errmsg);
}
#endif

/* ---------- */

static void *
client_gdbm_init (const GData *params)
{
        const char *filename;
        char *filename_user, *filename_chan, *filename_msg;
        GDBMInfo *info;

        filename = (const char *)g_datalist_get_data((GData **)&params, "filename");
#ifdef HAVE_GDBM
        filename_user = g_strdup_printf("%s-user.db", filename);
        filename_chan = g_strdup_printf("%s-chan.db", filename);
	filename_msg = g_strdup_printf("%s-msg.db", filename);
#else
	filename_user = g_strdup_printf("%s-user", filename);
	filename_chan = g_strdup_printf("%s-chan", filename);
	filename_msg = g_strdup_printf("%s-msg", filename);
#endif

        info = g_new0(GDBMInfo, 1);
        info->userdb = gdbm_open(filename_user, BLOCK_SIZE, GDBM_WRITER | GDBM_SYNC, 0600, &client_gdbm_fatal);
        info->chandb = gdbm_open(filename_chan, BLOCK_SIZE, GDBM_WRITER | GDBM_SYNC, 0600, &client_gdbm_fatal);
        info->msgdb = gdbm_open(filename_msg, BLOCK_SIZE, GDBM_WRITER | GDBM_SYNC, 0600, &client_gdbm_fatal);
        g_free(filename_user);
        g_free(filename_chan);
	g_free(filename_msg);

        if ((info->userdb == NULL) || (info->chandb == NULL) || (info->msgdb == NULL)) {
                g_free(info);
                return NULL;
        }
        return info;
}

static void
client_gdbm_shutdown (void *blob)
{
        GDBMInfo *info = (GDBMInfo *)blob;

        gdbm_close(info->userdb);
        gdbm_close(info->chandb);
	gdbm_close(info->msgdb);
        g_free(info);
}

static gboolean
client_gdbm_add_client (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = enaml_new ();
        client_update_block (client, block);
        enaml_add_long (block, "created", (long) time (NULL));
        ret = write_client_block (info, client, block, TRUE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_update_login (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }

        enaml_add_long (block, "last-login", (long) time (NULL));
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_update_logout (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }

        enaml_add_long (block, "last-logout", (long) time (NULL));
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_update_client (void *blob, const Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }

        client_update_block (client, block);
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_lookup_client (void *blob, const char *id, Client *client)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        const EnamlBlock *attr_block;
        const char *icon;

        if (client == NULL) {
                /* only wanna know if it exists */
                block = read_id_block (info, id, CLIENT_USER);
                if (block == NULL) {
                        return FALSE;
                }
                enaml_free (block);
                return TRUE;
        }

        block = read_id_block (info, id, client->flags);
        if (block == NULL) {
                return FALSE;
        }

        g_free (client->id);
        client->id = g_strdup (enaml_get_string (block, "id", NULL));
        g_free (client->nick);
        client->nick = g_strdup (client->id);
        g_free (client->email);
        client->email = g_strdup (enaml_get_string (block, "email", NULL));
        g_free (client->tagline);
        client->tagline = g_strdup (enaml_get_string (block, "tagline", NULL));
        client_string_to_prefs (client, enaml_get_string (block, "prefs", NULL));
        if (client->flags & CLIENT_USER) {
                client_string_to_privs (client, enaml_get_string (block, "privs", NULL));
        }

        client->iconsize = enaml_get_blob (block, "icon", &icon);
        if (icon != NULL) {
                client->icon = g_malloc (client->iconsize);
                memcpy (client->icon, icon, client->iconsize);
        }
        attr_block = enaml_get_block (block, "attributes");
        if (attr_block != NULL) {
                client->attributes = enaml_copy_block (attr_block);
        }

        enaml_free (block);

        return TRUE;
}

static gboolean
client_gdbm_check_password (void *blob, const char *id, const char *password, int flags)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        const char *c_password;
        gboolean ok;

        block = read_id_block (info, id, flags);
        if (block == NULL) {
                return FALSE;
        }
        c_password = enaml_get_string (block, "password", NULL);
        ok = FALSE;
        if (c_password != NULL) {
                if (strcmp (c_password, password) == 0) {
                        ok = TRUE;
                }
        }

        enaml_free (block);
        return ok;
}

static gboolean
client_gdbm_set_password (void *blob, const Client *client, const char *password)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;
        gboolean ret;

        block = read_client_block (info, client);
        if (block == NULL) {
                return FALSE;
        }
        enaml_add_string (block, "password", password, 0);
        ret = write_client_block (info, client, block, FALSE);
        enaml_free (block);
        return ret;
}

static gboolean
client_gdbm_get_seen (void *blob, const char *id, time_t *last_login, time_t *last_logout, int flags)
{
        GDBMInfo *info = (GDBMInfo *)blob;
        EnamlBlock *block;

        block = read_id_block (info, id, flags);
        if (block == NULL) {
                return FALSE;
        }
        *last_login = (time_t) enaml_get_long (block, "last-login", 0L);
        *last_logout = (time_t) enaml_get_long (block, "last-logout", 0L);
        enaml_free (block);
        return TRUE;
}

static gboolean
client_gdbm_get_messages (void *blob, const Client *client, GList **messages)
{
	GDBMInfo *info = (GDBMInfo *)blob;
	datum key, value;
	char *id_lower;
	char **strv;
	int i;

	id_lower = g_strdup(client->id);
	g_strdown(id_lower);
	key.dptr = id_lower;
	key.dsize = strlen(id_lower)+1;
	value = gdbm_fetch(info->msgdb, key);

	if (value.dptr == NULL) {
		g_free(id_lower);
		return FALSE;
	}

	strv = g_strsplit(value.dptr, "\n", 0);
	*messages = NULL;
	for (i = 0; strv[i] != NULL; i++) {
		*messages = g_list_prepend(*messages, g_strdup(strv[i]));
	}
	g_strfreev(strv);
	*messages = g_list_reverse(*messages);

	free(value.dptr);

	/* now delete those messages */
	gdbm_delete(info->msgdb, key);
	g_free(id_lower);

	return TRUE;
}

static gboolean
client_gdbm_add_message (void *blob, const char *id, const char *message)
{
	GDBMInfo *info = (GDBMInfo *)blob;
	datum key, value;
	char *id_lower;
	char *newval;
	int ret;

	id_lower = g_strdup(id);
	g_strdown(id_lower);
	key.dptr = id_lower;
	key.dsize = strlen(id_lower)+1;
	value = gdbm_fetch(info->msgdb, key);

	newval = g_strdup_printf("%s%s%s", (value.dptr == NULL) ? "" : (char *)value.dptr,
				 (value.dptr == NULL) ? "" : "\n",
				 message);
	if (value.dptr != NULL) {
		free(value.dptr);
	}
	value.dptr = newval;
	value.dsize = strlen(newval)+1;
	ret = gdbm_store(info->msgdb, key, value, GDBM_REPLACE);
	g_free(newval);
	g_free(id_lower);

	return (ret == 0);
}

DBModule *
client_gdbm_get_module (void)
{
        DBModule *module;

        module = g_new0 (DBModule, 1);
        module->init = &client_gdbm_init;
        module->shutdown = &client_gdbm_shutdown;
        module->add_client = &client_gdbm_add_client;
        module->update_login = &client_gdbm_update_login;
        module->update_logout = &client_gdbm_update_logout;
        module->update_client = &client_gdbm_update_client;
        module->lookup_client = &client_gdbm_lookup_client;
        module->check_password = &client_gdbm_check_password;
        module->set_password = &client_gdbm_set_password;
        module->get_seen = &client_gdbm_get_seen;
	module->get_messages = &client_gdbm_get_messages;
	module->add_message = &client_gdbm_add_message;
        module->empty = NULL;

        return module;
}
